`timescale 1ps / 1ps


module FlashWrapper#(
                      parameter ADDR_X_WIDTH = 4,
                      parameter ADDR_Y_WIDTH = 3,
                      parameter DOUT_WIDTH   = 8
                    ) 
                    ( 
                      input  logic                                 arst,
                      input  logic                                 controlFlash_0r0d,
                      input  logic                                 controlFlash_0r1d,
                      output logic                                 controlFlash_a,
                      input  logic                                 SE_Req,
                      output logic [DOUT_WIDTH-1:0]                SE_0a0d,
                      output logic [DOUT_WIDTH-1:0]                SE_0a1d,
                      input  logic [ADDR_X_WIDTH+ADDR_Y_WIDTH-1:0] adr_0r0d,
                      input  logic [ADDR_X_WIDTH+ADDR_Y_WIDTH-1:0] adr_0r1d,
                      output logic                                 adr_Ack,
                      input  logic [3:0][DOUT_WIDTH-1:0]           dataRead_0r0d,
                      input  logic [3:0][DOUT_WIDTH-1:0]           dataRead_0r1d,
                      output logic [3:0]                           dataRead_a
                    );


  logic [ADDR_X_WIDTH+ADDR_Y_WIDTH-1:0] adr_Data;
  logic                                 IFREN;
  logic  [1:0]                          REDEN;
  logic  [ADDR_X_WIDTH-1:0]             XADR;
  logic                                 XE;
  logic  [ADDR_Y_WIDTH-1:0]             YADR;
  logic                                 YE;
  logic                                 PV;
  logic                                 EV;
  logic                                 SE;
  logic                                 controlFlash;
  logic                                 dataValid;
  logic [ADDR_X_WIDTH+ADDR_Y_WIDTH-1:0] DOUT;
  logic  [DOUT_WIDTH*4-1:0]             dataRead;



  Flash    #(
            .ADDR_X_WIDTH  (ADDR_X_WIDTH), 
            .ADDR_Y_WIDTH  (ADDR_Y_WIDTH), 
            .DOUT_WIDTH    (DOUT_WIDTH ) 
             )
    u_Flash (
             .IFREN      (IFREN),
             .REDEN      (REDEN),
             .XADR       (XADR),
             .XE         (XE),
             .YADR       (YADR),
             .YE         (YE),
             .PV         (PV),
             .EV         (EV),
             .SE         (SE_Req),
             .DATAVALID  (dataValid),
             .DOUT       (DOUT)
             );



  DualRailLatch latch_controlFlash (.arst(arst), .S(controlFlash_0r1d), .R(controlFlash_0r0d), .Q(controlFlash), .Qn(), .ack(controlFlash_a));

 

  always_comb begin
    if (controlFlash == 1) begin
      IFREN = 0;
      REDEN = 0;
      XADR  = adr_Data[ADDR_X_WIDTH-1:0];
      XE    = 1'b1;
      YADR  = adr_Data[ADDR_X_WIDTH+ADDR_Y_WIDTH-1:ADDR_X_WIDTH];
      YE    = 1'b1;
      PV    = 0;
      EV    = 0;
      SE    = SE_Req;
    end 
    else begin
      IFREN = 0;
      REDEN = 0;
      XADR  = 0;
      XE    = 1'b0;
      YADR  = 0;
      YE    = 1'b0;
      PV    = 0;
      EV    = 0;
      SE    = 0;
    end
  end



  DualRailLatch #(.BITS(ADDR_X_WIDTH+ADDR_Y_WIDTH))
    latch_adr(.arst(arst), .S(adr_0r1d), .R(adr_0r0d), .Q(adr_Data), .Qn(), .ack(adr_Ack));


  assign SE_Ack = SE_Req & dataValid; 
  
  assign SE_0a1d = {ADDR_X_WIDTH+ADDR_Y_WIDTH{SE_Req & dataValid}} & DOUT;
  assign SE_0a0d = {ADDR_X_WIDTH+ADDR_Y_WIDTH{SE_Req & dataValid}} & ~DOUT;


  DualRailLatch #(.BITS(ADDR_X_WIDTH+ADDR_Y_WIDTH))
    latch_dataRead_0(.arst(arst), .S(dataRead_0r1d[0]), .R(dataRead_0r0d[0]), .Q(dataRead[7:0]),   .Qn(), .ack(dataRead_a[0]));

  DualRailLatch #(.BITS(ADDR_X_WIDTH+ADDR_Y_WIDTH))
    latch_dataRead_1(.arst(arst), .S(dataRead_0r1d[1]), .R(dataRead_0r0d[1]), .Q(dataRead[15:8]),  .Qn(), .ack(dataRead_a[1]));

  DualRailLatch #(.BITS(ADDR_X_WIDTH+ADDR_Y_WIDTH))
    latch_dataRead_2(.arst(arst), .S(dataRead_0r1d[2]), .R(dataRead_0r0d[2]), .Q(dataRead[23:16]), .Qn(), .ack(dataRead_a[2]));

  DualRailLatch #(.BITS(ADDR_X_WIDTH+ADDR_Y_WIDTH))
    latch_dataRead_3(.arst(arst), .S(dataRead_0r1d[3]), .R(dataRead_0r0d[3]), .Q(dataRead[31:24]), .Qn(), .ack(dataRead_a[3]));

endmodule


module DualRailLatch #(parameter BITS = 1)
                      (
                      input  logic arst,
                      input  logic [BITS-1:0] S,
                      input  logic [BITS-1:0] R,
                      output logic [BITS-1:0] Q,
                      output logic [BITS-1:0] Qn,
                      output logic            ack
                      );

  logic [BITS-1:0] ack_int;

  genvar i;
  generate
    for (i=0 ; i<BITS ; i++) begin
    
      logic Qint;
      logic Qint_n;

      always_comb
        if(R[i] | arst) begin
          Qint   = 1'b0;
          Qint_n = 1'b1;
        end
        else if (S[i]) begin
          Qint   = 1'b1;
          Qint_n = 1'b0;
        end
    
      assign #1ns Q[i]  = Qint;
      assign #1ns Qn[i] = Qint_n;
      assign #1ns ack_int[i] = S[i] & Q[i] | R[i] & Qn[i];
    end
  endgenerate
  
  assign ack = &ack_int;

endmodule





