`timescale 1ps / 1ps


module Flash#(
              parameter ADDR_X_WIDTH = 4,
              parameter ADDR_Y_WIDTH = 3,
              parameter DOUT_WIDTH   = 8
             )
             (
              input                     IFREN,
              input [1:0]               REDEN,
              input [ADDR_X_WIDTH-1:0]  XADR,
              input                     XE,
              input [ADDR_Y_WIDTH-1:0]  YADR,
              input                     YE,
              input                     PV,
              input                     EV,
              input                     SE,
              output                    DATAVALID,
              output [DOUT_WIDTH-1:0]   DOUT
             );


  time Tdh  = 0.5ns;
  time Tacc = 30ns;
  time Tpws = 5ns;
  
  logic [ADDR_Y_WIDTH + ADDR_X_WIDTH -1 :0]                         addr;
  logic [DOUT_WIDTH-1:0]                                            dataAtAddr;
  logic [2**ADDR_Y_WIDTH-1:0][2**ADDR_X_WIDTH-1:0][DOUT_WIDTH-1:0]  mem;

  logic                                                             valid;
  logic                                                             dataValid;
  logic [DOUT_WIDTH-1:0]                                            dOut; 

  time                                                              SEset;

  logic                                                             inputValid;


  assign addr = {YADR & {ADDR_Y_WIDTH{YE}}, XADR & {ADDR_X_WIDTH{XE}} };
  assign DATAVALID = dataValid;
  assign DOUT      = dOut;

  initial begin
    $timeformat(-6,3," us",0);

    dataValid = 1'b0;
    valid     = 1'b0;
    for (int y = 0 ; y<(2**ADDR_Y_WIDTH) ; y++) begin
      for (int x = 0 ; x<(2**ADDR_X_WIDTH) ; x++) begin
        mem[y][x]       = $urandom;
      end
    end
  end

  assign inputValid =  IFREN !== 1'bX && IFREN !== 1'bz && 
                       ^REDEN !== 1'bx &&
                       PV !== 1'bX && PV !== 1'bZ && 
                       EV !== 1'bX && EV !== 1'bZ && 
                       XE !== 1'bX && XE !== 1'bZ && 
                       YE !== 1'bX && YE !== 1'bZ && 
                       ^XADR  !== 1'bx &&
                       ^YADR  !== 1'bx &&
                       SE !== 1'bX && SE !== 1'bZ;

  always begin
    @(addr, XE, YE)
      if (!inputValid) begin         // Give out X if any of the control signals are X
        #Tdh dataAtAddr = {DOUT_WIDTH{1'bX}};
        valid = 1'b0;
      end
      else if (XE === 1'b1 && YE === 1'b1) begin
        #Tdh dataAtAddr = mem[YADR][XADR];
        valid = 1'b1;
      end
      else begin    // else if (XE === 1'b0 || YE === 1'b0)                       // Give X out if addresses not enabled
        #Tdh dataAtAddr = {DOUT_WIDTH{1'bX}};
        valid = 1'b0;
      end
  end


  always begin
    fork
      begin
        @(posedge SE)
          #Tacc dOut = dataAtAddr;
          dataValid  = valid;
      end
      begin
        @dataAtAddr
          dOut = {DOUT_WIDTH{1'bX}};
          dataValid  = 1'b0;
      end
    join_any;
    disable fork;
  end


  always begin
    @(posedge SE)
      SEset = $time;
    @(negedge SE)
      assert ($time - SEset >= Tpws)
        //$display("%t: SE long enough: %t", $time, $time - SEset);
      else begin
        $error("%t: SE not long enough. Must be minimum %t, but is %t", $time, Tpws, $time - SEset);
        dOut = {DOUT_WIDTH{1'bX}};
      end
  end

endmodule
